## This file generates a number of p-values across different parameter values, so it
## can be used for local tests in the parameter space.

suppressPackageStartupMessages(library(argparse))
source(file = 'regression_test_functions.R')
data_loc <- "~/icml_tests/data/"
#data_loc <- '~/Desktop/icml_tests/data/'


parser <- ArgumentParser()
parser$add_argument("-n", "--nsim", type="integer", default=20, 
                    help="Sample sizes for samples S1 and S2 [default %(default)s]",
                    metavar="number")
parser$add_argument("-r", "--rep", type="integer", default=20, 
                    help="Number of repetitions [default %(default)s]",
                    metavar="number")
parser$add_argument("-b", "--bootstrap", type="integer", default=100, 
                    help="Number of bootstrap rounds [default %(default)s]",
                    metavar="number")
parser$add_argument("-d", "--dim", type="integer", default=50, 
                    help="Maximum dimension reached by simulation [default %(default)s]",
                    metavar="number")
parser$add_argument("-l", "--length-theta", type="integer", default=10, 
                    help="Number of theta values tested between -5 and 5 [default %(default)s]",
                    metavar="number")
parser$add_argument("-m", "--model", default="random_forest", 
                    help="Model type - currently either log_reg or random_forest [default %(default)s]")

args <- parser$parse_args()
n_sim <- args$nsim
d <- args$dim
length_out <- args$length_theta
n_rep <- args$rep
n_bootstrap <- args$bootstrap

if (args$model == 'log_reg'){
  model_func <- regstat
} else {
  model_func <- rf_fit
  model <- 'random_forest'
}

theta_vector <- rep(seq(0, 1, length.out = length_out), n_rep)
d_vector <- rep(d, n_rep*length_out)
bootstrap_vector <- rep(n_bootstrap, n_rep*length_out)
nsim_vector <- rep(n_sim, n_rep*length_out)

list_results <- c()
for (i in 1:length(theta_vector)){
  print(paste0('Working on iteration ', i, ' of ', length(theta_vector)))
  list_results[[i]] <- local_test_function_bernoulli(n=nsim_vector[i], d=d_vector[i], 
                                                   theta=theta_vector[i], 
                                                   model_func=model_func,
                                                   n_bootstrap = bootstrap_vector[i])
}

mat_test_results <- do.call(rbind, list_results)
mat_results <- cbind(theta_vector, d_vector, bootstrap_vector,nsim_vector,
                     mat_test_results)
colnames(mat_results) <- c('theta','dimension', 'n_bootstrap', 'n_sim', 
                           'regression_test', 'energy_test', 'mmd_test')


out_list <- list()
out_list[[1]] <- mat_results
filename_out <- paste0(data_loc, 'test_results_localtest_bernoulli_n', n_sim, '_dim', d, '_lenout', 
                       length_out, '_model', model, 'bootstrap', n_bootstrap, '.RData')
save(out_list, file=filename_out)


############## Temp Plots for Assessing Performances Locally ####################

plot_df <- out_list[[1]] %>% as.data.frame() %>%
  group_by(theta) %>% summarise(power_regression = mean(regression_test < 0.05),
                                power_et = mean(energy_test < 0.05),
                                power_mdd = mean(mmd_test < 0.05)) %>%
  as.data.frame()
rep <- dim(out_list[[1]] %>% as.data.frame())[1]/length_out

local_test_plot_df <- reshape2::melt(plot_df, id = c('theta'))
title_plot_local <- paste0('Local Test, n=', n_sim, ', ',
                           ', Dim:', d, ', Boostrap Number: ', n_bootstrap, 
                           ', Reps: ', rep)
ggplot(local_test_plot_df, aes(x=theta, y=value, color=variable)) +
  geom_line(size=1.5) + theme_bw(base_size = 8) + 
  labs(color="Test", x="Theta", y="Power", title=title_plot_local) +
  scale_color_hue(labels = c("Our GoF Test", "Energy Test", "MMD Test")) +
  ylim(c(0,1))

