# This a fork of the following Github Repositoryhttps://github.com/ikostrikov/pytorch-flows
# re-adapted for running a MAF with the CAMELUS data

import argparse
import copy
import math
import sys

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
from tqdm import tqdm
from tensorboardX import SummaryWriter
from datetime import datetime
from sklearn.model_selection import train_test_split

import datasets
import flows as fnn
from utils import repeat, ilmun_test


# Training settings
parser = argparse.ArgumentParser(description='PyTorch Flows')
parser.add_argument(
    '--batch-size',
    type=int,
    default=100,
    help='input batch size for training (default: 100)')
parser.add_argument(
    '--test-batch-size',
    type=int,
    default=1000,
    help='input batch size for testing (default: 1000)')
parser.add_argument(
    '--epochs',
    type=int,
    default=1000,
    help='number of epochs to train (default: 1000)')
parser.add_argument(
    '--lr', type=float, default=0.0001, help='learning rate (default: 0.0001)')
parser.add_argument(
    '--dataset',
    default='LINC',
    help='POWER | GAS | HEPMASS | MINIBONE | BSDS300 | MOONS')
parser.add_argument(
    '--flow', default='maf', help='flow to use: maf | realnvp | glow')
parser.add_argument(
    '--no-cuda',
    action='store_true',
    default=False,
    help='disables CUDA training')
parser.add_argument(
    '--num-blocks',
    type=int,
    default=5,
    help='number of invertible blocks (default: 5)')
parser.add_argument(
    '--seed', type=int, default=1, help='random seed (default: 1)')
parser.add_argument(
    '--log-interval',
    type=int,
    default=1000,
    help='how many batches to wait before logging training status')
parser.add_argument(
    '--cond',
    action='store_true',
    default=False,
    help='train class conditional flow')
parser.add_argument(
    '--all_bins',
    action='store_true',
    default=False,
    help='Consider all bins in training or not')
parser.add_argument(
    '--num_hidden',
    type=int,
    default=1024,
    help='how many batches to wait before logging training status')
parser.add_argument(
    '--train',
    type=int,
    default=1024,
    help='how many batches to wait before logging training status')

args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()
device = torch.device("cuda:0" if args.cuda else "cpu")
cond = True

torch.manual_seed(args.seed)
if args.cuda:
    torch.cuda.manual_seed(args.seed)

kwargs = {'num_workers': 4, 'pin_memory': True} if args.cuda else {}
all_bins_str = 'all_bins' if args.all_bins else ''
dir_data = 'dataCSV/%s' % (all_bins_str + '/')
if args.train == 1000:
    training_data_flnm = dir_data + 'sim_training_data_subsampled55.csv'
    training_params_flnm = dir_data + 'sim_params_training_data_subsampled55_ohe.npy'
    testing_data_flnm = dir_data + 'sim_testing_data_subsampled55.csv'
    testing_params_flnm = dir_data + 'sim_params_testing_data_subsampled55_ohe.npy'
elif args.train == 5000:
    training_data_flnm = dir_data + 'sim_training_data_subsampled5_5000training.csv'
    training_params_flnm = dir_data + 'sim_params_training_data_subsampled5_5000training_ohe.npy'
    testing_data_flnm = dir_data + 'sim_testing_data_subsampled5_5000training.csv'
    testing_params_flnm = dir_data + 'sim_params_testing_data_subsampled5_5000training_ohe.npy'
elif args.train == 500:
    training_data_flnm = dir_data + 'sim_training_data_subsampled5_500training.csv'
    training_params_flnm = dir_data + 'sim_params_training_data_subsampled5_500training_ohe.npy'
    testing_data_flnm = dir_data + 'sim_testing_data_subsampled5_500training.csv'
    testing_params_flnm = dir_data + 'sim_params_testing_data_subsampled5_500training_ohe.npy'
elif args.train == 200:
    training_data_flnm = dir_data + 'sim_training_data_subsampled5_200training.csv'
    training_params_flnm = dir_data + 'sim_params_training_data_subsampled5_200training_ohe.npy'
    testing_data_flnm = dir_data + 'sim_testing_data_subsampled5_200training.csv'
    testing_params_flnm = dir_data + 'sim_params_testing_data_subsampled5_200training_ohe.npy'
else:
    raise OSError('No Training Files for Requested Training Size: %s' % args.train)

print(args.cond)

## Import and Setup Data

if args.cond:
    x_train_full = pd.read_csv(training_data_flnm).set_index('Unnamed: 0').values
    y_train_full = np.load(training_params_flnm)

    x_train_data, x_val_data, y_train_data, y_val_data = train_test_split(
        x_train_full, y_train_full, test_size=0.1, random_state=7)

    train_tensor = torch.from_numpy(x_train_data)
    train_labels = torch.from_numpy(y_train_data)
    train_dataset = torch.utils.data.TensorDataset(train_tensor, train_labels)

    val_tensor = torch.from_numpy(x_val_data)
    val_labels = torch.from_numpy(y_val_data)
    val_dataset = torch.utils.data.TensorDataset(val_tensor, val_labels)

    x_test_data = pd.read_csv(testing_data_flnm).set_index('Unnamed: 0').values
    y_test_data = np.load(testing_params_flnm)
    test_tensor = torch.from_numpy(x_test_data)
    test_labels = torch.from_numpy(y_test_data)
    test_dataset = torch.utils.data.TensorDataset(test_tensor, test_labels)
    num_cond_inputs = 50
else:
    x_train_full = pd.read_csv(training_data_flnm).set_index('Unnamed: 0').values
    y_train_full = np.load(training_params_flnm)

    print(x_train_full.shape, y_train_full.shape)

    x_train_data, x_val_data, y_train_data, y_val_data = train_test_split(
        x_train_full, y_train_full, test_size=0.1, random_state=7)

    train_tensor = torch.from_numpy(x_train_data)
    train_dataset = torch.utils.data.TensorDataset(train_tensor)

    val_tensor = torch.from_numpy(x_val_data)
    val_dataset = torch.utils.data.TensorDataset(val_tensor)

    x_test_data = pd.read_csv(testing_data_flnm).set_index('Unnamed: 0').values
    test_tensor = torch.from_numpy(x_test_data)
    test_dataset = torch.utils.data.TensorDataset(test_tensor)
    num_cond_inputs = None

train_loader = torch.utils.data.DataLoader(
    train_dataset, batch_size=args.batch_size, shuffle=True, **kwargs)

valid_loader = torch.utils.data.DataLoader(
    val_dataset,
    batch_size=args.test_batch_size,
    shuffle=False,
    drop_last=False,
    **kwargs)

test_loader = torch.utils.data.DataLoader(
    test_dataset,
    batch_size=args.test_batch_size,
    shuffle=False,
    drop_last=False,
    **kwargs)

num_inputs = 13 if args.all_bins else 7
num_hidden = args.num_hidden
act = 'tanh'

modules = []

assert args.flow in ['maf', 'maf-split', 'maf-split-glow', 'realnvp', 'glow']
if args.flow == 'glow':
    mask = torch.arange(0, num_inputs) % 2
    mask = mask.to(device).float()

    print("Warning: Results for GLOW are not as good as for MAF yet.")
    for _ in range(args.num_blocks):
        modules += [
            fnn.BatchNormFlow(num_inputs),
            fnn.LUInvertibleMM(num_inputs),
            fnn.CouplingLayer(
                num_inputs, num_hidden, mask, num_cond_inputs,
                s_act='tanh', t_act='relu')
        ]
    mask = 1 - mask
elif args.flow == 'realnvp':
    mask = torch.arange(0, num_inputs) % 2
    mask = mask.to(device).float()

    for _ in range(args.num_blocks):
        modules += [
            fnn.CouplingLayer(
                num_inputs, num_hidden, mask, num_cond_inputs,
                s_act='tanh', t_act='relu'),
            fnn.BatchNormFlow(num_inputs)
        ]
        mask = 1 - mask
elif args.flow == 'maf':
    for _ in range(args.num_blocks):
        modules += [
            fnn.MADE(num_inputs, num_hidden, num_cond_inputs, act=act),
            fnn.BatchNormFlow(num_inputs),
            fnn.Reverse(num_inputs)
        ]
elif args.flow == 'maf-split':
    for _ in range(args.num_blocks):
        modules += [
            fnn.MADESplit(num_inputs, num_hidden, num_cond_inputs,
                          s_act='tanh', t_act='relu'),
            fnn.BatchNormFlow(num_inputs),
            fnn.Reverse(num_inputs)
        ]
elif args.flow == 'maf-split-glow':
    for _ in range(args.num_blocks):
        modules += [
            fnn.MADESplit(num_inputs, num_hidden, num_cond_inputs,
                          s_act='tanh', t_act='relu'),
            fnn.BatchNormFlow(num_inputs),
            fnn.InvertibleMM(num_inputs)
        ]

model = fnn.FlowSequential(*modules)

for module in model.modules():
    if isinstance(module, nn.Linear):
        nn.init.orthogonal_(module.weight)
        if hasattr(module, 'bias') and module.bias is not None:
            module.bias.data.fill_(0)

model.to(device)
optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=1e-6)
writer = SummaryWriter(comment=args.flow + "_" + args.dataset)
global_step = 0


def train(epoch):
    global global_step, writer
    model.train()
    train_loss = 0

    pbar = tqdm(total=len(train_loader.dataset))
    for batch_idx, data in enumerate(train_loader):
        if isinstance(data, list):
            if len(data) > 1:
                cond_data = data[1].float()
                cond_data = cond_data.to(device).float()
            else:
                cond_data = None

            data = data[0]
        data = data.to(device)
        optimizer.zero_grad()
        loss = -model.log_probs(data.float(), cond_data).mean()
        train_loss += loss.item()
        loss.backward()
        optimizer.step()

        pbar.update(data.size(0))
        pbar.set_description('Train, Log likelihood in nats: {:.6f}'.format(
            -train_loss / (batch_idx + 1)))

        writer.add_scalar('training/loss', loss.item(), global_step)
        global_step += 1

    pbar.close()

    for module in model.modules():
        if isinstance(module, fnn.BatchNormFlow):
            module.momentum = 0

    if args.cond:
        with torch.no_grad():
            model(train_loader.dataset.tensors[0].to(data.device).float(),
                  train_loader.dataset.tensors[1].to(data.device).float())
    else:
        with torch.no_grad():
            model(train_loader.dataset.tensors[0].to(data.device).float())

    for module in model.modules():
        if isinstance(module, fnn.BatchNormFlow):
            module.momentum = 1


def validate(epoch, model, loader, prefix='Validation', run_pvals=True):
    global global_step, writer

    model.eval()
    val_loss = 0

    pbar = tqdm(total=len(loader.dataset))
    pbar.set_description('Eval')
    for batch_idx, data in enumerate(loader):
        if isinstance(data, list):
            if len(data) > 1:
                cond_data = data[1].float()
                cond_data = cond_data.to(device).float()
            else:
                cond_data = None

            data = data[0]
        data = data.to(device)
        with torch.no_grad():
            val_loss += -model.log_probs(data.float(), cond_data).sum().item()  # sum up batch loss
        pbar.update(data.size(0))
        pbar.set_description('Val, Log likelihood in nats: {:.6f}'.format(
            -val_loss / pbar.n))

    writer.add_scalar('validation/LL', val_loss / len(loader.dataset), epoch)

    pbar.close()

    # Evaluate 10 p_values
    if args.cond and run_pvals and epoch % 10 == 0:
        model.eval()
        batch_size = 10
        rep = 200
        fixed_noise = torch.Tensor(batch_size * rep, num_inputs).normal_()
        y = torch.arange(batch_size).unsqueeze(-1) % 50
        y_onehot = torch.FloatTensor(batch_size, 50)
        y_onehot.zero_()
        y_onehot.scatter_(1, y, 1)
        y_onehot = repeat(y_onehot, [rep, 1])

        with torch.no_grad():
            out = model.sample(batch_size, noise=fixed_noise, cond_inputs=y_onehot).detach().cpu().data.numpy()

        out_p_vals = []
        pbar = tqdm(total=batch_size, desc="Calculate P-values")
        for i in range(batch_size):
            test_temp = x_test_data[i * rep: (i + 1) * rep, :].astype(int)
            comp_temp = out[i * rep: (i + 1) * rep, :].astype(int)

            out_p_vals.append(ilmun_test(sample=test_temp, generated=comp_temp, d=num_inputs))
            pbar.update(1)
        print(out_p_vals)

    return val_loss / len(loader.dataset)


best_validation_loss = float('inf')
best_validation_epoch = 0
best_model = model

for epoch in range(args.epochs):
    print('\nEpoch: {}'.format(epoch))

    train(epoch)
    validation_loss = validate(epoch, model, valid_loader)

    if epoch - best_validation_epoch >= 10:
        break

    if validation_loss < best_validation_loss:
        best_validation_epoch = epoch
        best_validation_loss = validation_loss
        best_model = copy.deepcopy(model)

    print(
        'Best validation at epoch {}: Average Log Likelihood in nats: {:.4f}'.
            format(best_validation_epoch, -best_validation_loss))

validate(best_validation_epoch, best_model, test_loader, prefix='Test', run_pvals=False)

flnm_out = 'runs/models/%s_model_%strainpoint_%shidden_%sepochs_%sblocks_%strainsamples_%s_%s_%s.pt' % (
    args.flow, train_tensor.shape[0], num_hidden, args.epochs, args.num_blocks, args.train,
    'cond' if args.cond else 'nocond', 'all_bins' if args.all_bins else '',
    datetime.strftime(datetime.today(), '%Y-%m-%d-%H-%M')
)
torch.save(best_model.state_dict(), flnm_out)
